from dataclasses import dataclass
import yaml
import os


@dataclass
class Network:
    """
    The network class contains network settings for the VmClient.
    """

    chain_id: str
    chain_grpc_endpoint: str
    nilvm_grpc_endpoint: str

    @classmethod
    def devnet(
        cls,
        nilvm_grpc_endpoint: str,
        chain_grpc_endpoint: str,
    ) -> "Network":
        """
        Initializes a network configuration compatible with a Nillion devnet.

        By default the devnet starts without any SSL configuration so all of the `tls_*` parameters are not required.

        Arguments
        ---------
        nilvm_grpc_endpoint
            The Nillion network bootnode endpoint.

        chain_grpc_endpoint
            The nilchain gRPC endpoint.

        Example
        -------

        .. code-block:: py3

            config = Network.devnet(
                nilvm_grpc_endpoint="http://127.0.0.1:37939",
                chain_grpc_endpoint="localhost:26649",
            )

        """
        return cls(
            chain_id="nillion-chain-devnet",
            chain_grpc_endpoint=chain_grpc_endpoint,
            nilvm_grpc_endpoint=nilvm_grpc_endpoint,
        )

    @classmethod
    def from_config(cls, network_name: str) -> "Network":
        """
        Load a network configuration from the filesystem.

        This looks up a network configuration under `~/.config/nillion/networks`. This allows easily loading
        pre-existing network configurations, like the one dumped by `nillion-devnet` when it starts.

        Arguments
        ---------
        network_name
            The name of the network to be loaded.

        Example
        -------

        .. code-block:: py3

            config = Network.from_config("devnet")
        """
        config_home = _config_path_root()
        devnet_path = os.path.join(config_home, f"nillion/networks/{network_name}.yaml")
        try:
            with open(devnet_path) as fd:
                config = yaml.safe_load(fd)
        except Exception as ex:
            raise NetworkConfigLoadError(f"could not load configuration: {ex}")

        payments_config = config["payments"]
        nilvm_grpc_endpoint = config["bootnode"]
        chain_grpc_endpoint = payments_config["nilchain_grpc_endpoint"]
        chain_id = payments_config["nilchain_chain_id"]
        if not chain_grpc_endpoint:
            raise NetworkConfigLoadError("chain gRPC endpoint not set")
        if not chain_id:
            raise NetworkConfigLoadError("chain id")
        return cls(
            chain_id=chain_id,
            nilvm_grpc_endpoint=nilvm_grpc_endpoint,
            chain_grpc_endpoint=chain_grpc_endpoint,
        )


def _config_path_root() -> str:
    config_home = os.getenv("XDG_CONFIG_HOME")
    if not config_home:
        home = os.getenv("HOME")
        if not home:
            raise NetworkConfigLoadError("could not find configurations directory")
        config_home = os.path.join(home, ".config")
    return config_home


class NetworkConfigLoadError(Exception):
    """
    An exception raised when loading a network configuration.
    """

    pass



"""
Nillion client.
"""

from secp256k1 import PrivateKey
from .ids import UserId, ProgramId, ValuesId, ComputeId
from .network import Network
from .vm_operation import Operation, InputPartyBinding, OutputPartyBinding
from .permissions import Permissions, PermissionsDelta
from .client import VmClient, PermissionDeniedError, InternalError, NotFoundError
from .payer import NilChainPayer, Payer
from nillion_client_core import (
    Integer,
    SecretInteger,
    UnsignedInteger,
    SecretUnsignedInteger,
    Array,
    Boolean,
    SecretBoolean,
    SecretBlob,
    EcdsaPrivateKey,
    EcdsaDigestMessage,
    EcdsaSignature,
)
from nillion_client_proto.nillion.preprocessing.v1.element import PreprocessingElement
from cosmpy.crypto.keypairs import PrivateKey as NilChainPrivateKey

__all__ = [
    "UserId",
    "ProgramId",
    "ValuesId",
    "ComputeId",
    "Network",
    "Operation",
    "InputPartyBinding",
    "OutputPartyBinding",
    "Permissions",
    "PermissionsDelta",
    "VmClient",
    "PermissionDeniedError",
    "InternalError",
    "NotFoundError",
    "NilChainPayer",
    "Payer",
    "Integer",
    "SecretInteger",
    "UnsignedInteger",
    "SecretUnsignedInteger",
    "Array",
    "Boolean",
    "SecretBoolean",
    "SecretBlob",
    "EcdsaPrivateKey",
    "EcdsaDigestMessage",
    "EcdsaSignature",
    "PreprocessingElement",
    "PrivateKey",
    "NilChainPrivateKey",
]



import os

import pytest
import nillion_client

import secp256k1

from nillion_client.ids import UserId
from nillion_client.network import Network
from nillion_client.permissions import (
    PermissionCommand,
    Permissions,
    PermissionsDelta,
)
from grpclib import GRPCError


def relative_to_current_file(relative_path):
    """
    Convert a relative path to one relative to the current file's directory.
    """
    base_path = os.path.dirname(__file__)
    return os.path.normpath(os.path.join(base_path, relative_path))


@pytest.fixture(scope="session", autouse=True)
def devnet_setup():
    try:
        homedir = os.getenv("HOME")
        config_file_path = f"{homedir}/.config/nillion/nillion-devnet.env"
        grpc_endpoint, nilchain_private_key, nilchain_grpc_endpoint = None, None, None
        with open(config_file_path, "r") as config_file:
            for line in config_file:
                if "NILLION_GRPC_ENDPOINT" in line:
                    grpc_endpoint = line.split("=")[1].strip()
                if "NILLION_NILCHAIN_PRIVATE_KEY_0" in line:
                    nilchain_private_key = line.split("=")[1].strip()
                if "NILLION_NILCHAIN_GRPC" in line:
                    nilchain_grpc_endpoint = line.split("=")[1].strip()

        if not grpc_endpoint or not nilchain_private_key or not nilchain_grpc_endpoint:
            raise RuntimeError("Failed to read Nillion devnet config file")

        yield (
            nilchain_grpc_endpoint,
            grpc_endpoint,
            nilchain_private_key,
        )

    except Exception as e:
        print(f"Failed to start Nillion devnet: {e}")
        raise


async def new_client(devnet_setup) -> nillion_client.VmClient:
    (
        nilchain_grpc_endpoint,
        grpc_endpoint,
        nilchain_private_key,
    ) = devnet_setup

    signing_key = secp256k1.PrivateKey()

    network = nillion_client.Network.devnet(
        nilvm_grpc_endpoint=grpc_endpoint,
        chain_grpc_endpoint=nilchain_grpc_endpoint,
    )

    chain_client = nillion_client.NilChainPayer(
        network,
        wallet_private_key=nillion_client.NilChainPrivateKey(
            bytes.fromhex(nilchain_private_key)
        ),
        gas_limit=10000000,
    )
    vm_client = await nillion_client.VmClient.create(signing_key, network, chain_client)

    return vm_client


@pytest.mark.asyncio
async def test_pool_status(devnet_setup):
    """Test that we can fetch the pool status"""

    client = await new_client(devnet_setup)

    results = await client.pool_status().invoke()

    assert any(
        result.element == nillion_client.PreprocessingElement.LAMBDA
        for result in results
    ), "No lambda element found in pool"

    client.close()


@pytest.mark.asyncio
async def test_store_retrieve_all_value_types(devnet_setup):
    """Test that we can store and retrieve values"""

    client = await new_client(devnet_setup)

    values = {
        "int": nillion_client.Integer(42),
        "sint": nillion_client.SecretInteger(43),
        "uint": nillion_client.UnsignedInteger(43),
        "suint": nillion_client.SecretUnsignedInteger(43),
        "bool": nillion_client.Boolean(True),
        "sbool": nillion_client.SecretBoolean(False),
        "sblob": nillion_client.SecretBlob(bytearray("1234", "utf-8")),
        "array": nillion_client.Array(
            [nillion_client.Integer(1), nillion_client.Integer(2)]
        ),
        "key": nillion_client.EcdsaPrivateKey(bytearray(os.urandom(32))),
        "message": nillion_client.EcdsaDigestMessage(bytearray(os.urandom(32))),
        "signature": nillion_client.EcdsaSignature(
            (bytearray([1, 2, 3]), bytearray([1, 2, 3]))
        ),
    }

    values_id = await client.store_values(values, 1).invoke()
    returned_values = await client.retrieve_values(values_id).invoke()

    assert returned_values == values

    client.close()


@pytest.mark.asyncio
async def test_update_values(devnet_setup):
    """Test that we can store and retrieve values"""

    client = await new_client(devnet_setup)

    values = {
        "foo": nillion_client.Integer(42),
    }

    values_id = await client.store_values(values, 1).invoke()
    updated_values = {
        "bar": nillion_client.SecretBoolean(True),
    }

    identifier = await client.store_values(
        updated_values, ttl_days=1, update_identifier=values_id
    ).invoke()
    assert identifier == values_id

    returned_values = await client.retrieve_values(values_id).invoke()

    assert returned_values == updated_values

    client.close()


@pytest.mark.asyncio
async def test_delete_values(devnet_setup):
    """Test that we can store and delete values"""

    client = await new_client(devnet_setup)

    # Store a value, then delete it
    values = {
        "value1": nillion_client.Integer(42),
        "value2": nillion_client.SecretInteger(43),
    }

    values_id = await client.store_values(values, 1).invoke()

    await client.delete_values(values_id).invoke()

    # Check that retrieving the value fails
    with pytest.raises(GRPCError) as e:
        await client.retrieve_values(values_id).invoke()
    assert "not found" in str(e.value)

    # Check that deleting the value again fails
    with pytest.raises(GRPCError) as e:
        await client.delete_values(values_id).invoke()
    assert "not found" in str(e.value)

    client.close()


@pytest.mark.asyncio
async def test_store_values_retrieve_overwrite_permissions(devnet_setup):
    """Test that we can store values and retrieve their permissions"""

    client = await new_client(devnet_setup)

    signing_key = secp256k1.PrivateKey()
    other_user_id = UserId.from_public_key(signing_key.pubkey)  # type: ignore

    permissions = nillion_client.Permissions.defaults_for_user(client.user_id)
    permissions.allow_retrieve(other_user_id)

    # Check that we can retrieve permissions after storing values
    values = {
        "value1": nillion_client.Integer(42),
        "value2": nillion_client.SecretInteger(43),
    }

    values_id = await client.store_values(values, 1, permissions=permissions).invoke()

    returned_permissions = await client.retrieve_permissions(values_id).invoke()

    assert returned_permissions == permissions

    # Check we can update permissions
    permissions.allow_compute(other_user_id, nillion_client.ProgramId("dummyProgramId"))

    await client.overwrite_permissions(values_id, permissions).invoke()

    returned_permissions = await client.retrieve_permissions(values_id).invoke()

    assert returned_permissions == permissions

    client.close()


@pytest.mark.asyncio
async def test_update_permissions(devnet_setup):
    """Test that we can store values and retrieve their permissions"""

    client = await new_client(devnet_setup)
    signing_key = secp256k1.PrivateKey()
    other_user_id = UserId.from_public_key(signing_key.pubkey)  # type: ignore

    values = {
        "value1": nillion_client.Integer(42),
        "value2": nillion_client.SecretInteger(43),
    }
    values_id = await client.store_values(values, ttl_days=1).invoke()

    delta = PermissionsDelta(retrieve=PermissionCommand(grant=set([other_user_id])))
    await client.update_permissions(values_id, delta).invoke()

    permissions = await client.retrieve_permissions(values_id).invoke()
    assert other_user_id in permissions.retrieve

    client.close()


@pytest.mark.asyncio
async def test_basic_compute(devnet_setup):
    """Test that we can store and compute a program"""

    client = await new_client(devnet_setup)

    test_program = relative_to_current_file("resources/programs/main.nada.bin")
    program = open(test_program, "rb").read()

    program_id = await client.store_program("main", program).invoke()

    values = {
        "my_int1": nillion_client.SecretInteger(40),
        "my_int2": nillion_client.SecretInteger(2),
    }

    compute_id = await client.compute(
        program_id,
        input_bindings=[
            nillion_client.InputPartyBinding(party_name="Party1", user=client.user_id)
        ],
        output_bindings=[
            nillion_client.OutputPartyBinding(
                party_name="Party1", users=[client.user_id]
            )
        ],
        values=values,
    ).invoke()

    results = await client.retrieve_compute_results(compute_id).invoke()

    assert results == {"sum": nillion_client.SecretInteger(42)}

    client.close()


@pytest.mark.asyncio
async def test_complex_compute(devnet_setup):
    client_party1 = await new_client(devnet_setup)
    client_party2 = await new_client(devnet_setup)
    client_output = await new_client(devnet_setup)

    test_program = relative_to_current_file("resources/programs/main_complex.nada.bin")
    program = open(test_program, "rb").read()

    program_id = await client_party1.store_program("main", program).invoke()

    values_p2 = {
        "my_int2": nillion_client.SecretInteger(2),
    }
    permissions = Permissions(client_party2.user_id)
    permissions.allow_compute(client_party1.user_id, program_id)
    values_p2_id = await client_party2.store_values(
        values_p2, ttl_days=1, permissions=permissions
    ).invoke()

    values_p1 = {
        "my_int1": nillion_client.SecretInteger(40),
    }

    compute_id = await client_party1.compute(
        program_id,
        input_bindings=[
            nillion_client.InputPartyBinding(
                party_name="Party1", user=client_party1.user_id
            ),
            nillion_client.InputPartyBinding(
                party_name="Party2", user=client_party2.user_id
            ),
        ],
        output_bindings=[
            nillion_client.OutputPartyBinding(
                party_name="Party3", users=[client_output.user_id]
            )
        ],
        values=values_p1,
        value_ids=[values_p2_id],
    ).invoke()

    results = await client_output.retrieve_compute_results(compute_id).invoke()

    assert results == {"sum": nillion_client.SecretInteger(42)}

    client_party1.close()
    client_party2.close()
    client_output.close()


def test_network_config():
    # simply load it to ensure it doesn't throw
    Network.from_config("devnet")


@pytest.mark.asyncio
async def test_store_program(devnet_setup):
    """Test that we can store and compute a program"""

    client = await new_client(devnet_setup)

    test_program = relative_to_current_file("resources/programs/main.nada.bin")
    program = open(test_program, "rb").read()

    await client.store_program(
        "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890+.:_-",
        program,
    ).invoke()
    client.close()


@pytest.mark.asyncio
async def test_invalid_program_name(devnet_setup):
    client = await new_client(devnet_setup)
    test_program = relative_to_current_file("resources/programs/main.nada.bin")
    program = open(test_program, "rb").read()
    with pytest.raises(Exception):
        await client.store_program("main/nope", program).invoke()
    client.close()


@pytest.mark.asyncio
async def test_use_balance(devnet_setup):
    client = await new_client(devnet_setup)
    balance = await client.balance()
    assert balance.balance == 0

    # add some funds
    amount = 100
    await client.add_funds(amount)

    # ensure our balance went up
    balance = await client.balance()
    assert balance.balance == amount

    # run an operation and ensure it went down
    await client.pool_status().invoke()
    balance = await client.balance()
    assert balance.balance < amount

    client.close()



import asyncio
import logging
import enum
import hashlib
import ssl
import secrets
from typing import List, Dict, Optional, Mapping
from urllib.parse import urlparse

from grpclib.events import SendRequest, listen
from grpclib.client import Channel
from betterproto.lib.google.protobuf import Empty
import secp256k1
from nillion_client_proto.nillion.compute.v1.invoke import (
    InvokeComputeRequest,
)
from nillion_client_proto.nillion.compute.v1.retrieve import (
    RetrieveResultsRequest,
    RetrieveResultsResponse,
)
from nillion_client_proto.nillion.compute.v1 import ComputeStub
from nillion_client_proto.nillion.leader_queries.v1.pool_status import (
    PreprocessingOffsets,
    PoolStatusRequest,
)
from nillion_client_proto.nillion.leader_queries.v1 import LeaderQueriesStub
from nillion_client_proto.nillion.membership.v1.cluster import (
    Cluster,
    ClusterMember,
    Prime,
)
from nillion_client_proto.nillion.membership.v1 import MembershipStub
from nillion_client_proto.nillion.payments.v1.quote import (
    SignedQuote,
    PriceQuoteRequest,
    StoreValues,
    InvokeCompute,
    RetrieveValues,
    PreprocessingRequirement,
    StoreProgram,
    RetrievePermissions,
    OverwritePermissions,
    UpdatePermissions,
    ProgramMetadata,
)
from nillion_client_proto.nillion.payments.v1.balance import (
    AccountBalanceResponse,
    AddFundsPayload,
    AddFundsRequest,
)
from nillion_client_proto.nillion.payments.v1.receipt import (
    PaymentReceiptRequest,
    SignedReceipt,
)
from nillion_client_proto.nillion.payments.v1 import PaymentsStub
from nillion_client_proto.nillion.permissions.v1.retrieve import (
    RetrievePermissionsRequest,
)
from nillion_client_proto.nillion.permissions.v1 import PermissionsStub
from nillion_client_proto.nillion.permissions.v1.overwrite import (
    OverwritePermissionsRequest,
)
from nillion_client_proto.nillion.permissions.v1.update import UpdatePermissionsRequest
from nillion_client_proto.nillion.preprocessing.v1.element import PreprocessingElement
from nillion_client_proto.nillion.programs.v1 import ProgramsStub
from nillion_client_proto.nillion.programs.v1.store import StoreProgramRequest
from nillion_client_proto.nillion.values.v1.delete import DeleteValuesRequest
from nillion_client_proto.nillion.values.v1.retrieve import RetrieveValuesRequest
from nillion_client_proto.nillion.values.v1 import ValuesStub
from nillion_client_proto.nillion.values.v1.store import StoreValuesRequest
from nillion_client_core import (
    EncryptedNadaValues,
    PartyId,
    SecretMasker,
    encode_values,
    PartyJar,
    decode_values,
    compute_values_size,
    extract_program_metadata,
)

from .ids import UserId, ValuesId, ProgramId, ComputeId
from .network import Network
from .auth import AuthInterceptor
from .payer import Payer
from .permissions import Permissions, PermissionsDelta
from .vm_operation import (
    Operation,
    StoreValuesOperation,
    OverwritePermissionsOperation,
    RetrievePermissionsOperation,
    RetrieveValuesOperation,
    DeleteValuesOperation,
    StoreProgramOperation,
    ComputeOperation,
    RetrieveComputeResultsOperation,
    PoolStatusOperation,
    PayableOperation,
    InputPartyBinding,
    OutputPartyBinding,
    UpdatePermissionsOperation,
    NadaValue,
)
from .client_retry import _invoke_with_retry


class InternalError(Exception):
    """Exception raised for internal errors in the library."""


class NotFoundError(Exception):
    """Exception raised when values are not found."""


class PermissionDeniedError(Exception):
    """Exception raised a permission to run an operation has been denied."""


class PaymentMode(enum.Enum):
    """The payment mode to be used"""

    """
    Pay per operation, making a payment on nilchain every time an operation is ran.
    """
    PAY_PER_OPERATION = enum.auto()

    """
    Pay from the balance in the nilvm network, if any. If not enough balance is available, this will fall back
    to a payment per operation.
    """
    FROM_BALANCE = enum.auto()


class _ClusterMember:
    def __init__(self, member: ClusterMember, channel: Channel, party_id: PartyId):
        self.member = member
        self.channel = channel
        self.party_id = party_id
        self.values_service = ValuesStub(channel)
        self.permissions_service = PermissionsStub(channel)
        self.programs_service = ProgramsStub(channel)
        self.compute_service = ComputeStub(channel)


class _ClusterMemberLeader(_ClusterMember):
    def __init__(self, member: ClusterMember, channel: Channel, party_id: PartyId):
        super().__init__(member, channel, party_id)
        self.leader_queries_service = LeaderQueriesStub(channel)
        self.payment_service = PaymentsStub(channel)


def _sha256(data: bytes):
    sha256_hash = hashlib.sha256()
    sha256_hash.update(data)
    return sha256_hash.digest()


def _check_responses_equals(responses):
    if not responses:
        raise RuntimeError("Expected a responses list")

    first_element = responses[0]

    if all(element == first_element for element in responses[1:]):
        return first_element

    raise ValueError(f"Expected all nodes to return the same result: {responses}")


class VmClient:
    """
    A class to interact with the Nillion network.

    This class allows performing all operations on the Nillion network, such as storing and retrieving secrets,
    uploading programs, invoking computations, etc.

    Example
    -------

    .. code-block:: py3

        import os
        from nillion_client import PrivateKey, Network, NilChainPayer, NilChainPrivateKey, VmClient

        # The private key that will represent the identity of the user performing actions in the network.
        private_key = PrivateKey()

        # Load the config dumped by the `nillion-devnet` automatically on start and use it as the network.
        network = Network.from_config("devnet")

        # The payer that will be used to pay for operations in the network.
        nilchain_private_key = os.getenv("NILLION_NILCHAIN_KEY")
        chain_client = NilChainPayer(
            network,
            wallet_private_key=NilChainPrivateKey(bytes.fromhex(nilchain_private_key)),
            gas_limit=10000000,
        )

        # Finally, create the client
        client = await VmClient.create(private_key, network, payer)
    """

    user_id: UserId
    """The user identifier associated with this client."""

    network: Network
    """The network used by this client."""

    cluster: Cluster
    """The cluster definition that this client is using."""

    payment_mode: PaymentMode
    """The payment mode to be used."""

    def __init__(
        self,
        key: secp256k1.PrivateKey,
        network: Network,
        payer: Payer,
        payment_mode: PaymentMode,
        _raise_if_called=True,
    ):
        """
        Constructs a new instance of the client.

        Users should not directly use this function and should use the :meth:`VmClient.create` function instead.

        :meta private:
        """
        if _raise_if_called:
            raise Exception("VmClient must be initialized via VmClient.create")
        if not key.pubkey:
            raise ValueError("Key must have a public key")
        self._key = key
        self._members = []
        self._payer = payer
        self.payment_mode = payment_mode
        self.user_id = UserId.from_public_key(key.pubkey)
        self.network = network
        self.cluster = Cluster()

    def _create_channel(self, url: str, interceptors=None) -> Channel:
        grpc_endpoint_url = urlparse(url)
        use_tls = grpc_endpoint_url.scheme == "https"

        if use_tls:
            ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
            ssl_context.load_default_certs()
        else:
            ssl_context = None

        channel = Channel(
            host=grpc_endpoint_url.hostname,
            port=grpc_endpoint_url.port,
            ssl=ssl_context,
        )

        if interceptors:

            async def send_request_interceptors(event: SendRequest):
                for interceptor in interceptors:
                    await _invoke_with_retry(interceptor.intercept_send_request)(event)

            listen(channel, SendRequest, send_request_interceptors)

        return channel

    async def _async_init(self):
        bootnode_channel = self._create_channel(self.network.nilvm_grpc_endpoint)

        membership_service = MembershipStub(bootnode_channel)
        request = Empty()
        self.cluster = await _invoke_with_retry(membership_service.cluster)(request)

        parties = []

        # Create a token and a communication channel to each member
        for member in self.cluster.members:
            channel = self._create_channel(
                member.grpc_endpoint,
                interceptors=[AuthInterceptor(self._key, member.identity)],
            )
            party_id = PartyId.from_bytes(member.identity.contents)
            self._members.append(
                _ClusterMember(
                    member=member,
                    channel=channel,
                    party_id=party_id,
                )
            )
            parties.append(party_id)

        # The leader is currently part of the cluster but that won't necessarily always be the case.
        leader_channel = self._create_channel(
            self.cluster.leader.grpc_endpoint,
            interceptors=[AuthInterceptor(self._key, self.cluster.leader.identity)],
        )

        self.leader = _ClusterMemberLeader(
            member=self.cluster.leader,
            channel=leader_channel,
            party_id=PartyId.from_bytes(self.cluster.leader.identity.contents),
        )

        # Make sure the node actually supports in-network balances.
        if self.payment_mode == PaymentMode.FROM_BALANCE:
            version = await _invoke_with_retry(membership_service.node_version)(Empty())
            semver = version.version
            # 0.8.0+ supports this and we consider no version to be devnet, which also supports it
            supports_balances = not semver or semver.major > 0 or version.minor >= 8
            if not supports_balances:
                logging.warning(
                    "Falling back to paying per operation as network does not support this"
                )
                self.payment_mode

        bootnode_channel.close()

        # Create a secret masker with a prime that corresponds to the cluster's.
        match self.cluster.prime:
            case Prime.SAFE_64_BITS:
                self.secret_masker = SecretMasker.new_64_bit_safe_prime(
                    self.cluster.polynomial_degree, parties
                )
            case Prime.SAFE_128_BITS:
                self.secret_masker = SecretMasker.new_128_bit_safe_prime(
                    self.cluster.polynomial_degree, parties
                )
            case Prime.SAFE_256_BITS:
                self.secret_masker = SecretMasker.new_256_bit_safe_prime(
                    self.cluster.polynomial_degree, parties
                )

    @classmethod
    async def create(
        cls,
        key: secp256k1.PrivateKey,
        network: Network,
        payer: Payer,
        payment_mode: PaymentMode = PaymentMode.FROM_BALANCE,
    ) -> "VmClient":
        """
        Create a new Nillion client.

        Arguments
        ---------
        key
            The private key that will represent the client's identity in the network.
        network
            The network the client should connect to.
        payer
            The payer that will pay for all operations performed in the network.

        Example
        -------

        .. code-block:: py3

            client = await VmClient.create(private_key, network, payer)
        """
        client = cls(key, network, payer, payment_mode, _raise_if_called=False)
        await client._async_init()
        return client

    async def request_quote(self, operation: Operation) -> SignedQuote:
        """
        Requests a quote for an operation.

        Users should generally not use this function directly and should instead use the concrete function
        for the action they're trying to perform, such as :meth:`VmClient.store_values`,
        """
        from . import vm_operation as operations

        if isinstance(operation, operations.StoreValuesOperation):
            classification = self.secret_masker.classify_values(operation.values)
            payload_size = compute_values_size(operation.values)
            request = PriceQuoteRequest(
                store_values=StoreValues(
                    particles_count=classification.particles,
                    secret_shared_count=classification.shares,
                    public_values_count=classification.public,
                    ttl_days=operation.ttl_days,
                    payload_size=payload_size,
                )
            )
        elif isinstance(operation, operations.ComputeOperation):
            payload_size = compute_values_size(operation.values)
            request = PriceQuoteRequest(
                invoke_compute=InvokeCompute(
                    program_id=operation.program_id,
                    values_payload_size=payload_size,
                )
            )
        elif isinstance(operation, operations.RetrieveValuesOperation):
            request = PriceQuoteRequest(
                retrieve_values=RetrieveValues(values_id=operation.values_id.bytes)
            )
        elif isinstance(operation, operations.StoreProgramOperation):
            program_metadata = extract_program_metadata(operation.program)
            program_size = len(operation.program)
            program_sha256 = _sha256(operation.program)

            preprocessing_requirements = [
                PreprocessingRequirement(
                    element=PreprocessingElement.ALPHA,
                    count=program_metadata.preprocessing_requirements.alpha_elements,
                )
            ]

            for (
                key,
                value,
            ) in program_metadata.preprocessing_requirements.runtime_elements.items():
                match key:
                    case "Compare":
                        element = PreprocessingElement.COMPARE
                    case "DivisionIntegerSecret":
                        element = PreprocessingElement.DIVISION_SECRET_DIVISOR
                    case "EqualsIntegerSecret":
                        element = PreprocessingElement.EQUALITY_SECRET_OUTPUT
                    case "Lambda":
                        element = PreprocessingElement.LAMBDA
                    case "Modulo":
                        element = PreprocessingElement.MODULO
                    case "ShareToParticle":
                        element = PreprocessingElement.SHARE_TO_PARTICLE
                    case "PublicOutputEquality":
                        element = PreprocessingElement.EQUALITY_PUBLIC_OUTPUT
                    case "TruncPr":
                        element = PreprocessingElement.TRUNC_PR
                    case "Trunc":
                        element = PreprocessingElement.TRUNC
                    case _:
                        raise InternalError(f"unknown preprocessing element: '{key}'")
                preprocessing_requirements.append(
                    PreprocessingRequirement(element=element, count=int(value))
                )

            request = PriceQuoteRequest(
                store_program=StoreProgram(
                    metadata=ProgramMetadata(
                        program_size=program_size,
                        memory_size=program_metadata.memory_size,
                        instruction_count=program_metadata.total_instructions,
                        instructions=program_metadata.instructions,
                        preprocessing_requirements=preprocessing_requirements,
                    ),
                    contents_sha256=program_sha256,
                    name=operation.program_name,
                )
            )
        elif isinstance(operation, operations.RetrievePermissionsOperation):
            request = PriceQuoteRequest(
                retrieve_permissions=RetrievePermissions(
                    values_id=operation.values_id.bytes
                )
            )
        elif isinstance(operation, operations.PoolStatusOperation):
            request = PriceQuoteRequest(pool_status=Empty())
        elif isinstance(operation, operations.OverwritePermissionsOperation):
            request = PriceQuoteRequest(
                overwrite_permissions=OverwritePermissions(
                    values_id=operation.values_id.bytes
                )
            )
        elif isinstance(operation, operations.UpdatePermissionsOperation):
            request = PriceQuoteRequest(
                update_permissions=UpdatePermissions(
                    values_id=operation.values_id.bytes
                )
            )
        else:
            raise InternalError(f"unknown operation: '{type(operation)}'")

        result = await _invoke_with_retry(self.leader.payment_service.price_quote)(
            request
        )
        return result

    def store_values(
        self,
        values: Mapping[str, NadaValue],
        ttl_days: int,
        permissions: Optional["Permissions"] = None,
        update_identifier: Optional[ValuesId] = None,
    ) -> PayableOperation[ValuesId]:
        """
        Store a set of values in the network.

        Any secret values will be masked automatically before uploading them.

        Arguments
        ---------
        values
            The values to store.

        ttl_days
            The number of days after which the values should be deleted. The higher this value, the higher the
            operation cost will be.

        permissions
            The permissions to be set for the uploaded values. By default only the uploader will have read and update
            permissions.

        update_identifier
            An identifier of the secret to be updated. If set, this turns this operation into an update.

        Returns
        -------
            An operation that once invoked returns an identifier that uniquely identifies the uploaded values.

        Example
        -------

        .. code-block:: py3

            values = {
                "foo": SecretInteger(42),
                "bar": Integer(1337),
            }
            await client.store_values(values, ttl_days=1).invoke()
        """
        return PayableOperation(
            StoreValuesOperation(
                self,
                values,
                ttl_days,
                permissions=permissions,
                update_identifier=update_identifier,
            )
        )

    def overwrite_permissions(
        self, values_id: ValuesId, permissions: Permissions
    ) -> PayableOperation[None]:
        """
        Overwrites the permissions on the given values id.

        This operation requires the user to have "update" permissions on the given values identifier.

        Arguments
        ---------
        values_id
            The identifier of the uploaded values for which permissions should be updated.

        permissions
            The permissions to be set.

        Returns
        -------
        An operation that once invoked will overwrite the permissions for the given values.

        Example
        -------

        .. code-block:: py3

            permissions = Permissions.default_for_user(user_id)
            await client.overwrite_permissions(values_id, permissions).invoke()
        """
        return OverwritePermissionsOperation(self, values_id, permissions).to_payable()

    def update_permissions(
        self, values_id: ValuesId, delta: PermissionsDelta
    ) -> PayableOperation[None]:
        """
        Updates the permissions on the given values id with the given delta. As opposed to
        :meth:`VmClient.overwrite_permissions`, this operation allows granting and revoking individual permissions
        without overwriting the entire set.

        This operation can only be invoked by the owner of the stored values.

        Arguments
        ---------
        values_id
            The identifier of the uploaded values for which permissions should be updated.

        delta
            The permissions to be granted and revoked.

        Returns
        -------
        An operation that once invoked will update the permissions for the given values.

        Example
        -------

        .. code-block:: py3

            # Grant permissions to retrieve these values to `other_user_id`.
            delta = PermissionsDelta(retrieve=PermissionCommand(grant=set([other_user_id])))
            await client.update_permissions(values_id, delta).invoke()
        """
        return UpdatePermissionsOperation(self, values_id, delta).to_payable()

    def retrieve_permissions(
        self, values_id: ValuesId
    ) -> PayableOperation[Permissions]:
        """
        Retrieves the permissions for the given values identifier.

        This operation requires the user to have "retrieve" permissions on the given values identifier.

        Arguments
        ---------
        values_id
            The identifier of the uploaded values for which permissions should be retrieved.

        Returns
        -------
            An operation that once invoked will retrieve the set of permissions currently associated with the values.

        Example
        -------

        .. code-block:: py3

            await client.retrieve_permissions(values_id).invoke()
        """
        return RetrievePermissionsOperation(self, values_id).to_payable()

    def retrieve_values(
        self, values_id: ValuesId
    ) -> PayableOperation[Dict[str, NadaValue]]:
        """
        Retrieves the values with the given identifier, performing unmasking to recover any secrets
        in the given set of values.

        This operation requires the user to have "retrieve" permissions on the given values identifier.

        Arguments
        ---------
        values_id
            The identifier of the uploaded values that should be retrieved.

        Returns
        -------
            An operation that once invoked will retrieve the stored values.

        Example
        -------

        .. code-block:: py3

            await client.retrieve_values(values_id).invoke()
        """
        return RetrieveValuesOperation(self, values_id).to_payable()

    def delete_values(self, values_id: ValuesId) -> DeleteValuesOperation:
        """
        Deletes the values with the given identifier.

        This operation requires the user to have "delete" permissions on the given values identifier.

        Arguments
        ---------
        values_id
            The identifier of the uploaded values that should be deleted.

        Returns
        -------
            An operation that once invoked will delete the stored values.

        Example
        -------

        .. code-block:: py3

            await client.delete_values(values_id).invoke()
        """
        return DeleteValuesOperation(self, values_id)

    def store_program(
        self, program_name: str, program: bytes
    ) -> PayableOperation[ProgramId]:
        """
        Stores a program in the network.

        Stored programs can by default be invoked by anyone, although their identifier is only known to the invoker
        and node operators.

        Arguments
        ---------
        program_name
            The name of the program being uploaded.

        program
            The contents of the `.nada.bin` file generated from an invocation to `pynadac`.

        Returns
        -------
            An operation that once invoked will return an identifier that can be used to reference this program
            in invocations to `compute` and when setting up values permissions for compute operations.

        Example
        -------

        .. code-block:: py3

            contents = open("/tmp/program.nada.bin", "rb").read()
            await client.store_program(program_name="my-test-program", program=contents).invoke()
        """
        return StoreProgramOperation(self, program_name, program).to_payable()

    def compute(
        self,
        program_id: ProgramId,
        input_bindings: List[InputPartyBinding],
        output_bindings: List[OutputPartyBinding],
        values: Mapping[str, NadaValue],
        value_ids: List[ValuesId] | None = None,
    ) -> PayableOperation[ComputeId]:
        """
        Invokes a computation.

        This operation returns immediately as soon as all initial validations for the program invocation are performed.

        The results for a computation should be fetched by output parties via the
        :meth:`VmClient.retrieve_compute_results` function.

        The name of the input and output parties must match the defined parties in the program being invoked.

        Arguments
        ---------
        program_id
            The identifier of the program being invoked.

        input_bindings
            The list of bindings that associate input parties in the program with Nillion user identifiers.

        output_bindings
            The list of bindings that associate output parties in the program with Nillion user identifiers.

        values
            The values to be used as compute time secrets. These values will only be used during the computation and
            will be discarded afterwards.

        value_ids
            The list of value identifiers to be used as inputs to this computation.

        Returns
        -------
            An operation that once invoked will return an identifier that can be used to fetch computation
            results via :meth:`VmClient.retrieve_compute_results`

        Example
        -------

        .. code-block:: py3

            values = {
                "foo": SecretInteger(40),
                "bar": SecretInteger(2),
            }

            # Invoke a program using the given input and output bindings. In this case we are the only party providing
            # inputs and we are the only party receiving outputs.
            compute_id = await client.compute(
                program_id,
                input_bindings=[
                    nillion_client.InputPartyBinding(party_name="Party1", user=client.user_id)
                ],
                output_bindings=[
                    nillion_client.OutputPartyBinding(
                        party_name="Party1", users=[client.user_id]
                    )
                ],
                values=values,
            ).invoke()

            # Now fetch the results.
            results = await client.retrieve_compute_results(compute_id).invoke()
        """
        return ComputeOperation(
            self, program_id, input_bindings, output_bindings, values, value_ids
        ).to_payable()

    def retrieve_compute_results(
        self, compute_id: ComputeId
    ) -> RetrieveComputeResultsOperation:
        """
        Retrieve the results of a compute operation.

        If the compute operation has finished, this will return the result immediately. If the operation is still
        ongoing, this will block until the operation finishes.

        The invoker user must have been set as an output binding in the :meth:`VmClient.compute` invocation. If the
        invoker was bound as an output party, it will only retrieve the subset of the outputs that belong to it
        based on the configured bindings.

        Arguments
        ---------
        compute_id
            The identifier of the compute instance to fetch the results for.

        Returns
        -------
            An operation that once invoked will return the result of the computation.

        Example
        -------

        .. code-block:: py3

            results = await client.retrieve_compute_results(compute_id).invoke()
        """
        return RetrieveComputeResultsOperation(self, compute_id)

    def pool_status(self) -> PayableOperation[List[PreprocessingOffsets]]:
        """
        Fetch the preprocessing pool status.

        Returns
        -------
            The available offsets for every preprocessing element.

        Example
        -------

        .. code-block:: py3

            offsets = await client.pool_status(compute_id)
        """
        return PoolStatusOperation(self).to_payable()

    async def get_payment_receipt(
        self, signed_quote: SignedQuote, tx_hash: str | None
    ) -> SignedReceipt:
        """
        Request to get a payment receipt for a paid operation.

        Arguments
        ---------
        signed_quote
            The quote to get a payment receipt for.

        tx_hash
            The transaction hash where the payment was made.

        Returns
        -------
            A signed receipt that can be used to prove to all nodes that the payment was made.
        """
        request = PaymentReceiptRequest(
            signed_quote=signed_quote, tx_hash=tx_hash or ""
        )
        result = await _invoke_with_retry(self.leader.payment_service.payment_receipt)(
            request
        )
        return result

    async def invoke_store_values(
        self,
        receipt: SignedReceipt,
        values: Mapping[str, NadaValue],
        permissions: Optional[Permissions] = None,
        update_identifier: Optional[ValuesId] = None,
    ) -> ValuesId:
        """
        Invoke a store values operation in the network.

        Arguments
        ---------
        receipt
            A receipt that proves the payment was made.

        values
            The values to store.

        ttl_days
            The number of days after which the values should be deleted. The higher this value, the higher the
            operation cost will be.

        permissions
            The permissions to be set for the uploaded values. By default only the uploader will have read and update
            permissions.

        update_identifier
            An identifier of the secret to be updated. If set, this turns this operation into an update.

        Returns
        -------
            The identifier for the uploaded values.

        Example
        -------

        .. code-block:: py3

            values = {
                "foo": SecretInteger(42),
                "bar": Integer(1337),
            }
            await client.store_values(receipt, values, ttl_days=1).invoke()

        .. note:: users should use generally use :meth:`VmClient.store_values` unless the API provided by that function
            doesn't satisfy their use cases.
        """
        encrypted_party_shares = self.secret_masker.mask(values)

        if permissions is None:
            permissions = Permissions.defaults_for_user(self.user_id)

        response_futs = []
        for member in self._members:
            encrypted_values = encrypted_party_shares.get(member.party_id)
            encoded_values = encode_values(encrypted_values)
            request = StoreValuesRequest(
                signed_receipt=receipt,
                bincode_values=encoded_values,
                permissions=permissions.to_proto(),
                update_identifier=update_identifier.bytes if update_identifier else b"",
            )
            response_futs.append(
                _invoke_with_retry(member.values_service.store_values)(request)
            )

        responses = await asyncio.gather(*response_futs)
        result = _check_responses_equals(responses)
        return ValuesId(bytes=result.values_id)

    async def invoke_overwrite_permissions(
        self, receipt: SignedReceipt, permissions: Permissions
    ):
        """
        Invokes an overwrite permissions operation for the given values id.

        Arguments
        ---------
        receipt
            A receipt that proves the payment was made.

        values_id
            The identifier of the uploaded values for which permissions should be updated.

        permissions
            The permissions to be set.

        Example
        -------

        .. code-block:: py3

            permissions = Permissions.default_for_user(user_id)
            await client.overwrite_permissions(receipt, values_id, permissions).invoke()

        .. note:: users should use generally use :meth:`VmClient.overwrite_permissions` unless the API provided by
            that function doesn't satisfy their use cases.
        """
        response_futs = []
        for member in self._members:
            request = OverwritePermissionsRequest(
                signed_receipt=receipt, permissions=permissions.to_proto()
            )
            response_futs.append(
                _invoke_with_retry(member.permissions_service.overwrite_permissions)(
                    request
                )
            )

        await asyncio.gather(*response_futs)

    async def invoke_update_permissions(
        self, receipt: SignedReceipt, delta: PermissionsDelta
    ):
        """
        Invokes an updates the permissions on the given values.

        Arguments
        ---------
        receipt
            A receipt that proves the payment was made.

        values_id
            The identifier of the uploaded values for which permissions should be updated.

        delta
            The permissions to be granted and revoked.

        Example
        -------

        .. code-block:: py3

            # Grant permissions to retrieve these values to `other_user_id`.
            delta = PermissionsDelta(retrieve=PermissionCommand(grant=set([other_user_id])))
            await client.update_permissions(receipt, values_id, delta).invoke()

        .. note:: users should use generally use :meth:`VmClient.update_permissions` unless the API provided by
            that function doesn't satisfy their use cases.
        """
        response_futs = []
        for member in self._members:
            request = UpdatePermissionsRequest(
                signed_receipt=receipt,
                retrieve=delta.retrieve.to_proto(),
                update=delta.update.to_proto(),
                delete=delta.delete.to_proto(),
                compute=delta.compute.to_proto(),
            )
            response_futs.append(
                _invoke_with_retry(member.permissions_service.update_permissions)(
                    request
                )
            )

        await asyncio.gather(*response_futs)

    async def invoke_retrieve_permissions(self, receipt: SignedReceipt) -> Permissions:
        """
        Invokes a retrieve permissions operation in the network.

        Arguments
        ---------
        receipt
            A receipt that proves the payment was made.

        values_id
            The identifier of the uploaded values for which permissions should be retrieved.

        Returns
        -------
            The set of permissions currently associated with the values.

        Example
        -------

        .. code-block:: py3

            await client.retrieve_permissions(receipt, values_id).invoke()

        .. note:: users should use generally use :meth:`VmClient.retrieve_permissions` unless the API provided by
            that function doesn't satisfy their use cases.
        """
        response_futs = []
        for member in self._members:
            request = RetrievePermissionsRequest(signed_receipt=receipt)
            response_futs.append(
                _invoke_with_retry(member.permissions_service.retrieve_permissions)(
                    request
                )
            )

        responses = await asyncio.gather(*response_futs)
        responses = [Permissions.from_proto(response) for response in responses]
        response = _check_responses_equals(responses)
        return response

    async def invoke_retrieve_values(
        self, receipt: SignedReceipt
    ) -> Dict[str, NadaValue]:
        """
        Invokes a retrieve values operation.

        Arguments
        ---------
        receipt
            A receipt that proves the payment was made.

        values_id
            The identifier of the uploaded values that should be retrieved.

        Returns
        -------
            The stored values.

        Example
        -------

        .. code-block:: py3

            await client.retrieve_values(receipt, values_id).invoke()

        .. note:: users should use generally use :meth:`VmClient.retrieve_values` unless the API provided by
            that function doesn't satisfy their use cases.
        """
        party_jar = PartyJar.new(len(self._members))
        response_futs = []

        async def retrieve_values(receipt, member):
            request = RetrieveValuesRequest(signed_receipt=receipt)
            response = await _invoke_with_retry(member.values_service.retrieve_values)(
                request
            )
            return member.party_id, response

        for member in self._members:
            response_futs.append(retrieve_values(receipt, member))

        responses = await asyncio.gather(*response_futs)

        for party_id, result in responses:
            encrypted_values = decode_values(result.bincode_values)
            party_jar.add_element(party_id, encrypted_values)

        return self.secret_masker.unmask(party_jar)

    async def invoke_delete_values(self, values_id: ValuesId) -> None:
        """
        Invoke a delete values operation in the network.

        Arguments
        ---------
        receipt
            A receipt that proves the payment was made.

        values_id
            The identifier of the uploaded values that should be deleted.

        Example
        -------

        .. code-block:: py3

            await client.delete_values(receipt, values_id).invoke()

        .. note:: users should use generally use :meth:`VmClient.delete_values` unless the API provided by
            that function doesn't satisfy their use cases.
        """
        response_futs = []
        for member in self._members:
            request = DeleteValuesRequest(values_id=values_id.bytes)
            response_futs.append(
                _invoke_with_retry(member.values_service.delete_values)(request)
            )

        await asyncio.gather(*response_futs)

    async def invoke_store_program(
        self, receipt: SignedReceipt, program: bytes
    ) -> ProgramId:
        """
        Invokes a store program operation in the network.

        Arguments
        ---------
        receipt
            A receipt that proves the payment was made.

        program_name
            The name of the program being uploaded.

        program
            The contents of the `.nada.bin` file generated from an invocation to `pynadac`.

        Returns
        -------
            The stored program's identifier.

        Example
        -------

        .. code-block:: py3

            contents = open("/tmp/program.nada.bin", "rb").read()
            await client.store_program(receipt, program_name="my-test-program", program=contents).invoke()

        .. note:: users should use generally use :meth:`VmClient.store_program` unless the API provided by
            that function doesn't satisfy their use cases.
        """
        response_futs = []
        for member in self._members:
            request = StoreProgramRequest(program=program, signed_receipt=receipt)
            response_futs.append(
                _invoke_with_retry(member.programs_service.store_program)(request)
            )

        responses = await asyncio.gather(*response_futs)
        result = _check_responses_equals(responses)
        return ProgramId(result.program_id)

    async def invoke_compute(
        self,
        receipt: SignedReceipt,
        input_bindings: List[InputPartyBinding],
        output_bindings: List[OutputPartyBinding],
        values: Mapping[str, NadaValue],
        value_ids: List[ValuesId] | None = None,
    ) -> ComputeId:
        """
        Invokes a compute operation in the network.

        Arguments
        ---------
        receipt
            A receipt that proves the payment was made.

        program_id
            The identifier of the program being invoked.

        input_bindings
            The list of bindings that associate input parties in the program with Nillion user identifiers.

        output_bindings
            The list of bindings that associate output parties in the program with Nillion user identifiers.

        values
            The values to be used as compute time secrets. These values will only be used during the computation and
            will be discarded afterwards.

        value_ids
            The list of value identifiers to be used as inputs to this computation.

        Returns
        -------
            An identifier for the compute operation.

        Example
        -------

        .. code-block:: py3

            values = {
                "foo": SecretInteger(40),
                "bar": SecretInteger(2),
            }

            # Invoke a program using the given input and output bindings. In this case we are the only party providing
            # inputs and we are the only party receiving outputs.
            compute_id = await client.compute(
                program_id,
                input_bindings=[
                    nillion_client.InputPartyBinding(party_name="Party1", user=client.user_id)
                ],
                output_bindings=[
                    nillion_client.OutputPartyBinding(
                        party_name="Party1", users=[client.user_id]
                    )
                ],
                values=values,
            ).invoke()

            # Now fetch the results.
            results = await client.retrieve_compute_results(receipt, compute_id).invoke()

        .. note:: users should use generally use :meth:`VmClient.compute` unless the API provided by
            that function doesn't satisfy their use cases.
        """
        encrypted_party_shares = self.secret_masker.mask(values)

        value_ids = value_ids or []
        response_futs = []
        for member in self._members:
            encrypted_values = encrypted_party_shares.get(member.party_id)
            encoded_values = encode_values(encrypted_values)
            request = InvokeComputeRequest(
                signed_receipt=receipt,
                value_ids=[values_id.bytes for values_id in value_ids],
                bincode_values=encoded_values,
                input_bindings=[binding.to_proto() for binding in input_bindings],
                output_bindings=[binding.to_proto() for binding in output_bindings],
            )
            response_futs.append(
                _invoke_with_retry(member.compute_service.invoke_compute)(request)
            )

        responses = await asyncio.gather(*response_futs)
        result = _check_responses_equals(responses)
        return ComputeId(bytes=result.compute_id)

    async def invoke_retrieve_compute_results(
        self, compute_id: ComputeId
    ) -> Dict[str, NadaValue]:
        """
        Invoke a retrieve compute results operation in the network.

        Arguments
        ---------
        receipt
            A receipt that proves the payment was made.

        compute_id
            The identifier of the compute instance to fetch the results for.

        Returns
        -------
            The result of the computation.

        Example
        -------

        .. code-block:: py3

            results = await client.retrieve_compute_results(receipt, compute_id).invoke()

        .. note:: users should use generally use :meth:`VmClient.retrieve_compute_results` unless the API provided by
            that function doesn't satisfy their use cases.
        """

        async def retrieve_results(
            compute_id: ComputeId, member: _ClusterMember
        ) -> tuple[PartyId, EncryptedNadaValues]:
            request = RetrieveResultsRequest(compute_id=compute_id.bytes)
            response_stream = _invoke_with_retry(
                member.compute_service.retrieve_results
            )(request)
            async for response in response_stream:
                match response:
                    case RetrieveResultsResponse(waiting_computation=_):
                        continue
                    case RetrieveResultsResponse(success=result):
                        encrypted_values = decode_values(result.bincode_values)
                        return member.party_id, encrypted_values
                    case RetrieveResultsResponse(error=error):
                        raise RuntimeError(f"Received error from cluster: {error}")
            raise Exception("Unreachable")

        results = await asyncio.gather(
            *[retrieve_results(compute_id, member) for member in self._members]
        )

        party_jar = PartyJar.new(len(self._members))

        for party_id, encrypted_values in results:
            party_jar.add_element(party_id, encrypted_values)

        return self.secret_masker.unmask(party_jar)

    async def invoke_pool_status(
        self, receipt: SignedReceipt
    ) -> List[PreprocessingOffsets]:
        """
        Invokes a preprocessing pool status operation in the network.

        Arguments
        ---------
        receipt
            A receipt that proves the payment was made.

        .. note:: users should use generally use :meth:`VmClient.pool_status` unless the API provided by
            that function doesn't satisfy their use cases.
        """

        request = PoolStatusRequest(signed_receipt=receipt)
        result = await _invoke_with_retry(
            self.leader.leader_queries_service.pool_status
        )(request)
        return result.offsets

    async def balance(self) -> AccountBalanceResponse:
        """
        Gets the balance associated with the user's account in the network.

        This balance will be preferred when running operations and can be topped up by calling
        :meth:`VmClient.add_funds`.
        """
        return await _invoke_with_retry(self.leader.payment_service.account_balance)(
            Empty()
        )

    async def add_funds(self, amount_unil: int, target_user: UserId | None = None):
        """
        Add funds to a user's account on the nillion network.

        By default this will fund the account tied to the user this client is currently using.

        Funds will be automatically used when performing payments unless the payment mode in the client is
        changed.

        .. note:: Funds will expire after 30 days regardless of use so don't add more funds than you intend to use
            in the short term
        """
        payments_config = await _invoke_with_retry(
            self.leader.payment_service.payments_config
        )(Empty())
        if amount_unil < payments_config.minimum_add_funds_payment:
            raise Exception(
                f"can't add less than {payments_config.minimum_add_funds_payment} unil"
            )
        nonce = secrets.token_bytes(32)
        payload = AddFundsPayload(
            recipient=(target_user or self.user_id).to_proto(), nonce=nonce
        ).SerializeToString()
        payload_hash = hashlib.sha256(payload).digest()
        tx_hash = await self._payer.submit_payment(amount_unil, payload_hash)
        request = AddFundsRequest(payload, tx_hash)
        await _invoke_with_retry(self.leader.payment_service.add_funds)(request)

    def close(self):
        """
        Closes the client and releases all resources associated with it.
        """
        for member in self._members:
            member.channel.close()
        self.leader.channel.close()

    def __del__(self):
        self.close()



import re
import logging
from grpclib import GRPCError, Status
from typing import Dict, Optional, List, Union, Generic, TypeVar, TYPE_CHECKING, Mapping
from dataclasses import dataclass

from nillion_client_core import (
    Integer,
    SecretInteger,
    SecretUnsignedInteger,
    Boolean,
    SecretBoolean,
    SecretBlob,
    Array,
    UnsignedInteger,
    EcdsaPrivateKey,
    EcdsaDigestMessage,
    EcdsaSignature,
)

from google.rpc.error_details_pb2 import PreconditionFailure  # type: ignore
from nillion_client_proto.nillion.payments.v1.quote import SignedQuote, PriceQuote
from nillion_client_proto.nillion.payments.v1.receipt import SignedReceipt
from nillion_client_proto.nillion.leader_queries.v1.pool_status import (
    PreprocessingOffsets,
)
from nillion_client_proto.nillion.compute.v1.invoke import (
    InputPartyBinding as ProtoInputPartyBinding,
    OutputPartyBinding as ProtoOutputPartyBinding,
)

from .ids import ValuesId, ProgramId, ComputeId, UserId

PROGRAM_NAME_REGEX = re.compile("[a-zA-Z0-9+.:_-]{1,128}")

if TYPE_CHECKING:
    from . import VmClient, Permissions, PermissionsDelta

NadaValue = Union[
    Integer,
    SecretInteger,
    SecretUnsignedInteger,
    Boolean,
    SecretBoolean,
    SecretBlob,
    Array,
    UnsignedInteger,
    EcdsaPrivateKey,
    EcdsaDigestMessage,
    EcdsaSignature,
]


@dataclass
class InputPartyBinding:
    """Represents the binding of a named input party in a program to a user id"""

    party_name: str
    user: UserId

    def to_proto(self) -> ProtoInputPartyBinding:
        """Converts an instance to its protobuf representation"""
        return ProtoInputPartyBinding(
            party_name=self.party_name, user=self.user.to_proto()
        )

    @staticmethod
    def from_proto(proto: ProtoInputPartyBinding) -> "InputPartyBinding":
        """Constructs an instance from its protobuf representation"""
        return InputPartyBinding(
            party_name=proto.party_name, user=UserId.from_proto(proto.user)
        )


@dataclass
class OutputPartyBinding:
    """Represents the binding of a named output party in a program to a user id"""

    party_name: str
    users: List[UserId]

    def to_proto(self) -> ProtoOutputPartyBinding:
        """Converts an instance to its protobuf representation"""
        return ProtoOutputPartyBinding(
            party_name=self.party_name, users=[user.to_proto() for user in self.users]
        )

    @staticmethod
    def from_proto(proto: ProtoOutputPartyBinding) -> "OutputPartyBinding":
        """Constructs an instance from its protobuf representation"""
        return OutputPartyBinding(
            party_name=proto.party_name,
            users=[UserId.from_proto(user) for user in proto.users],
        )


class StoreValuesOperation:
    def __init__(
        self,
        client: "VmClient",
        values: Mapping[str, NadaValue],
        ttl_days: int,
        permissions: Optional["Permissions"] = None,
        update_identifier: Optional[ValuesId] = None,
    ):
        self.client = client
        self.values = values
        self.ttl_days = ttl_days
        self.permissions = permissions
        self.update_identifier = update_identifier

    def to_payable(self) -> "PayableOperation[ValuesId]":
        return PayableOperation(self)

    async def invoke(self, receipt: SignedReceipt) -> ValuesId:
        """
        Invoke this operation.

        Arguments
        ---------
        receipt
            The receipt that proves the payment was made.

        Returns
        -------
            A unique identifier for the values stored.
        """
        return await self.client.invoke_store_values(
            receipt,
            self.values,
            permissions=self.permissions,
            update_identifier=self.update_identifier,
        )


class OverwritePermissionsOperation:
    def __init__(
        self, client: "VmClient", values_id: ValuesId, permissions: "Permissions"
    ):
        self.client = client
        self.values_id = values_id
        self.permissions = permissions

    def to_payable(self) -> "PayableOperation[None]":
        return PayableOperation(self)

    async def invoke(self, receipt: SignedReceipt) -> None:
        """
        Invoke this operation.

        Arguments
        ---------
        receipt
            The receipt that proves the payment was made.
        """
        return await self.client.invoke_overwrite_permissions(receipt, self.permissions)


class UpdatePermissionsOperation:
    def __init__(
        self, client: "VmClient", values_id: ValuesId, delta: "PermissionsDelta"
    ):
        self.client = client
        self.values_id = values_id
        self.delta = delta

    def to_payable(self) -> "PayableOperation[None]":
        return PayableOperation(self)

    async def invoke(self, receipt: SignedReceipt) -> None:
        """
        Invoke this operation.

        Arguments
        ---------
        receipt
            The receipt that proves the payment was made.
        """
        return await self.client.invoke_update_permissions(receipt, self.delta)


class RetrievePermissionsOperation:
    def __init__(self, client: "VmClient", values_id: ValuesId):
        self.client = client
        self.values_id = values_id

    def to_payable(self) -> "PayableOperation[Permissions]":
        return PayableOperation(self)

    async def invoke(self, receipt: SignedReceipt) -> "Permissions":
        """
        Invoke this operation.

        Arguments
        ---------
        receipt
            The receipt that proves the payment was made.

        Returns
        -------
            The permissions for the stored values.
        """
        return await self.client.invoke_retrieve_permissions(receipt)


class RetrieveValuesOperation:
    def __init__(self, client: "VmClient", values_id: ValuesId):
        self.client = client
        self.values_id = values_id

    def to_payable(self) -> "PayableOperation[Dict[str, NadaValue]]":
        return PayableOperation(self)

    async def invoke(self, receipt: SignedReceipt) -> Dict[str, NadaValue]:
        """
        Invoke this operation.

        Arguments
        ---------
        receipt
            The receipt that proves the payment was made.

        Returns
        -------
            The stored values in clear.
        """
        return await self.client.invoke_retrieve_values(receipt)


class DeleteValuesOperation:
    def __init__(self, client: "VmClient", values_id: ValuesId):
        self.client = client
        self.values_id = values_id

    async def invoke(self) -> None:
        """
        Invoke this operation.
        """
        return await self.client.invoke_delete_values(self.values_id)


class StoreProgramOperation:
    def __init__(self, client: "VmClient", program_name: str, program: bytes):
        self.client = client
        self.program_name = program_name
        self.program = program
        if not PROGRAM_NAME_REGEX.match(program_name):
            raise Exception("invalid program name")

    def to_payable(self) -> "PayableOperation[ProgramId]":
        return PayableOperation(self)

    async def invoke(self, receipt: SignedReceipt) -> ProgramId:
        """
        Invoke this operation.

        Arguments
        ---------
        receipt
            The receipt that proves the payment was made.

        Returns
        -------
            A unique identifier for the stored program.
        """
        return await self.client.invoke_store_program(receipt, self.program)


class ComputeOperation:
    def __init__(
        self,
        client: "VmClient",
        program_id: ProgramId,
        input_bindings: List[InputPartyBinding],
        output_bindings: List[OutputPartyBinding],
        values: Mapping[str, NadaValue],
        value_ids: List[ValuesId] | None = None,
    ):
        self.client = client
        self.program_id = program_id
        self.input_bindings = input_bindings
        self.output_bindings = output_bindings
        self.values = values
        self.value_ids = value_ids

    def to_payable(self) -> "PayableOperation[ComputeId]":
        return PayableOperation(self)

    async def invoke(self, receipt: SignedReceipt) -> ComputeId:
        """
        Invoke this operation.

        Arguments
        ---------
        receipt
            The receipt that proves the payment was made.

        Returns
        -------
            A unique identifier for this compute instance.
        """
        return await self.client.invoke_compute(
            receipt,
            self.input_bindings,
            self.output_bindings,
            self.values,
            self.value_ids,
        )


class RetrieveComputeResultsOperation:
    def __init__(self, client: "VmClient", compute_id: ComputeId):
        self.client = client
        self.compute_id = compute_id

    async def invoke(self) -> Dict[str, NadaValue]:
        """
        Invoke this operation.

        Returns
        -------
            The results of the computation.
        """
        return await self.client.invoke_retrieve_compute_results(self.compute_id)


class PoolStatusOperation:
    def __init__(self, client: "VmClient"):
        self.client = client

    def to_payable(self) -> "PayableOperation[List[PreprocessingOffsets]]":
        return PayableOperation(self)

    async def invoke(self, receipt: SignedReceipt) -> List[PreprocessingOffsets]:
        """
        Invoke this operation.

        Arguments
        ---------
        receipt
            The receipt that proves the payment was made.

        Returns
        -------
            The preprocessing pool offsets.
        """
        return await self.client.invoke_pool_status(receipt)


Operation = Union[
    StoreValuesOperation,
    OverwritePermissionsOperation,
    RetrievePermissionsOperation,
    RetrieveValuesOperation,
    DeleteValuesOperation,
    UpdatePermissionsOperation,
    StoreProgramOperation,
    ComputeOperation,
    RetrieveComputeResultsOperation,
    PoolStatusOperation,
]

T = TypeVar("T")


class PayableOperation(Generic[T]):
    def __init__(self, operation: Operation):
        self.operation = operation

    async def quote(self) -> "QuotedOperation[T]":
        """
        Get a price quote for this operation.
        """
        quote = await self.operation.client.request_quote(self.operation)
        return QuotedOperation(self.operation, quote)

    async def invoke(self) -> T:
        """
        Invoke this operation.

        This is the equivalent of calling

        .. code-block:: py3

            quoted_op = await operation.quote()
            paid_op = await quoted_op.pay()
            validated_op = await paid_op.validate()
            result = await validated_op.invoke()
        """
        quoted_op = await self.quote()
        paid_op = await quoted_op.pay()
        validated_op = await paid_op.validate()
        return await validated_op.invoke()


class QuotedOperation(Generic[T]):
    def __init__(self, operation: Operation, quote: SignedQuote):
        self.operation = operation
        self.quote = quote

    async def pay(self, *args, **kwargs) -> "PaidOperation[T]":
        """
        Pay for this operation.

        This will use the configured payer on the client to perform the payment.
        """
        from nillion_client.client import PaymentMode

        payment_mode = kwargs.pop("payment_mode", self.operation.client.payment_mode)
        if payment_mode == PaymentMode.PAY_PER_OPERATION:
            price_quote = PriceQuote()
            price_quote.parse(self.quote.quote)
            tx_hash = await self.operation.client._payer.submit_payment(
                price_quote.fees.total, price_quote.nonce, *args, **kwargs
            )
            return PaidOperation(self.operation, self.quote, tx_hash)
        elif payment_mode == PaymentMode.FROM_BALANCE:
            return PaidOperation(self.operation, self.quote, None)
        else:
            raise Exception("invalid payment mode")

    async def pay_and_invoke(self, *args, **kwargs) -> T:
        """
        Pay and invoke the operation.
        """
        paid_op = await self.pay(*args, **kwargs)
        validated_op = await paid_op.validate()
        return await validated_op.invoke()


class PaidOperation(Generic[T]):
    def __init__(self, operation: Operation, quote: SignedQuote, tx_hash: str | None):
        self.operation = operation
        self.quote = quote
        self.tx_hash = tx_hash

    async def validate(self) -> "ValidatedOperation[T]":
        try:
            receipt = await self.operation.client.get_payment_receipt(
                self.quote, self.tx_hash
            )
            return ValidatedOperation(self.operation, self.quote, receipt)
        except GRPCError as ex:
            if _is_balance_error(ex):
                from nillion_client.client import PaymentMode

                logging.warning(
                    "Not enough balance in network, making single time payment"
                )
                paid_operation = await QuotedOperation(self.operation, self.quote).pay(
                    payment_mode=PaymentMode.PAY_PER_OPERATION
                )
                return await paid_operation.validate()
            else:
                raise


class ValidatedOperation(Generic[T]):
    def __init__(
        self,
        operation: Operation,
        quote: SignedQuote,
        receipt: SignedReceipt,
    ):
        self.operation = operation
        self.quote = quote
        self.receipt = receipt

    async def invoke(self) -> T:
        """
        Invoke the operation.
        """
        return await self.operation.invoke(self.receipt)  # type: ignore


def _is_balance_error(e: GRPCError) -> bool:
    if e.status != Status.FAILED_PRECONDITION:
        return False
    for detail in e.details or []:
        if isinstance(detail, PreconditionFailure):
            for violation in detail.violations:
                if violation.type == "PAYMENT" and violation.subject == "BALANCE":
                    return True
    return False



import hashlib
import secp256k1
from nillion_client import Permissions
from nillion_client_proto.nillion.permissions.v1.permissions import (
    Permissions as ProtoPermissions,
    ComputePermissions as ProtoComputePermissions,
)

from nillion_client.ids import UserId
from nillion_client.permissions import (
    ComputePermission,
    ComputePermissionCommand,
    ComputePermissions,
    PermissionCommand,
)


def make_user_id(seed: str) -> UserId:
    raw_seed = hashlib.sha256(seed.encode("utf8")).digest()
    key = secp256k1.PrivateKey(raw_seed)
    return UserId.from_public_key(key.pubkey)  # type: ignore


def test_allow_retrieve():
    user = make_user_id("other")
    permissions = Permissions(owner=make_user_id("owner"))
    permissions.allow_retrieve(user)
    assert user in permissions.retrieve


def test_allow_delete():
    user = make_user_id("other")
    permissions = Permissions(owner=make_user_id("owner"))
    permissions.allow_delete(user)
    assert user in permissions.delete


def test_allow_update():
    user = make_user_id("other")
    permissions = Permissions(owner=make_user_id("owner"))
    permissions.allow_update(user)
    assert user in permissions.update


def test_allow_compute():
    user = make_user_id("user4")
    permissions = Permissions(owner=make_user_id("owner"))
    permissions.allow_compute(user, "program1")
    assert user in permissions.compute.permissions
    assert "program1" in permissions.compute.permissions[user].program_ids


def test_chained_methods():
    permissions = (
        Permissions(owner=make_user_id("owner"))
        .allow_retrieve(make_user_id("user1"))
        .allow_delete(make_user_id("user2"))
        .allow_update(make_user_id("user3"))
        .allow_compute(make_user_id("user4"), "program2")
    )
    assert make_user_id("user1") in permissions.retrieve
    assert make_user_id("user2") in permissions.delete
    assert make_user_id("user3") in permissions.update
    assert make_user_id("user4") in permissions.compute.permissions
    assert (
        "program2" in permissions.compute.permissions[make_user_id("user4")].program_ids
    )


def test_multiple_compute_permissions():
    user = make_user_id("other")
    permissions = (
        Permissions(owner=make_user_id("owner"))
        .allow_compute(user, "program3")
        .allow_compute(user, "program4")
    )
    user_permissions = permissions.compute.permissions[user]
    assert "program3" in user_permissions.program_ids
    assert "program4" in user_permissions.program_ids


def test_to_proto():
    owner = make_user_id("owner")
    user = make_user_id("other")
    permissions = Permissions(owner=owner)
    permissions.allow_retrieve(user)
    proto = permissions.to_proto()
    assert proto.owner.contents == owner.contents
    assert user.to_proto() in proto.retrieve


def test_from_proto():
    # Create a ProtoPermissions message for testing
    owner = make_user_id("owner")
    retrieve_user = make_user_id("retrieve")
    compute_user = make_user_id("compute")
    program_id = "program"
    proto = ProtoPermissions(
        owner=owner.to_proto(),
        retrieve=[retrieve_user.to_proto()],
        compute=[
            ProtoComputePermissions(
                user=compute_user.to_proto(), program_ids=[program_id]
            )
        ],
    )
    expected = (
        Permissions(owner=owner)
        .allow_retrieve(retrieve_user)
        .allow_compute(compute_user, program_id)
    )
    assert Permissions.from_proto(proto) == expected


def test_permissions_equality():
    user_id_1 = make_user_id("1")
    user_id_2 = make_user_id("2")

    # Create two identical Permissions objects
    permissions1 = Permissions.defaults_for_user(user_id_1)
    permissions2 = Permissions.defaults_for_user(user_id_1)

    assert (
        permissions1 == permissions2
    ), "Permissions objects should be equal when initialized identically"

    # Modify one Permissions object
    permissions2.allow_retrieve(user_id_2)

    assert (
        permissions1 != permissions2
    ), "Permissions objects should not be equal after modification"


def test_permission_command():
    command = PermissionCommand(
        grant=set([make_user_id("1"), make_user_id("2")]),
        revoke=set([make_user_id("1"), make_user_id("2")]),
    )
    decoded = PermissionCommand.from_proto(command.to_proto())
    assert decoded == command


def test_compute_permission_command():
    command = ComputePermissionCommand(
        grant=ComputePermissions(
            permissions={make_user_id("1"): ComputePermission(program_ids=set(["a"]))}
        ),
        revoke=ComputePermissions(
            permissions={make_user_id("2"): ComputePermission(program_ids=set(["b"]))}
        ),
    )
    decoded = ComputePermissionCommand.from_proto(command.to_proto())
    assert decoded == command



import datetime
import secrets

from nillion_client_proto.nillion.membership.v1.cluster import NodeId
import secp256k1
from grpclib.events import SendRequest
from nillion_client_proto.nillion.auth.v1.public_key import PublicKey, PublicKeyType
from nillion_client_proto.nillion.auth.v1.token import Token, SignedToken

HEADER_NAME = "x-nillion-token-bin"
_TOKEN_EXPIRY = 60
_RENEWAL_THRESHOLD = 10


class AuthInterceptor:
    """An interceptor that adds a token to authenticate the user on every request."""

    def __init__(self, signing_key: secp256k1.PrivateKey, identity: NodeId):
        self._signing_key = signing_key
        self._identity = identity
        self._token = _create_token(signing_key, identity)
        self._token_expiry_time = None
        self._renew_token()

    def _renew_token(self):
        self._token = _create_token(self._signing_key, self._identity)
        now = datetime.datetime.now(datetime.timezone.utc)
        self._token_expiry_time = now + datetime.timedelta(seconds=_TOKEN_EXPIRY)

    def _is_token_expired_or_about_to_expire(self):
        now = datetime.datetime.now(datetime.timezone.utc)
        return (
            self._token_expiry_time is None
            or (self._token_expiry_time - now).total_seconds() <= _RENEWAL_THRESHOLD
        )

    async def intercept_send_request(self, event: SendRequest):
        if self._is_token_expired_or_about_to_expire():
            self._renew_token()
        event.metadata[HEADER_NAME] = self._token


def _create_token(signing_key: secp256k1.PrivateKey, identity: NodeId):
    now = datetime.datetime.now(datetime.timezone.utc)
    expires_at = now + datetime.timedelta(seconds=_TOKEN_EXPIRY)

    token = Token(
        nonce=secrets.token_bytes(32),
        target_identity=identity,
        expires_at=expires_at,
    )

    public_key = PublicKey(
        key_type=PublicKeyType.SECP256K1,
        contents=signing_key.pubkey.serialize(),  # type: ignore
    )

    token = token.SerializeToString()
    signed_message = signing_key.ecdsa_sign(token)
    signature = signing_key.ecdsa_serialize_compact(signed_message)

    signed_token = SignedToken(
        serialized_token=token,
        public_key=public_key,
        signature=signature,
    )

    return signed_token.SerializeToString()



import hashlib
from uuid import UUID

from dataclasses import dataclass
import secp256k1

from nillion_client_proto.nillion.auth.v1.user import (
    UserId as ProtoUserId,
)

USER_ID_LENGTH = 20


@dataclass(eq=True, frozen=True)
class UserId:
    """
    A user identifier.

    User identifiers are derived from the public key used for authentication when performing operations in the
    network. User identifiers are non sensitive and can be shared with other users.
    """

    contents: bytes

    @staticmethod
    def parse(hex_bytes: str) -> "UserId":
        """
        Parse a user identifier from a hex encoded string.

        Arguments
        ---------
        hex_bytes
            The hex bytes that represent the user id.

        Returns
        -------
            The parsed user id.

        Example
        -------

        .. code-block:: py3

            user = UserId.parse("3113a1170de795e4b725b84d1e0b4cfd9ec58ce9")
        """
        try:
            contents = bytes.fromhex(hex_bytes)
        except Exception as ex:
            raise InvalidUserId(str(ex))
        if len(contents) != USER_ID_LENGTH:
            raise InvalidUserId(f"length should be {USER_ID_LENGTH} bytes")
        return UserId(contents)

    @staticmethod
    def from_public_key(public_key: secp256k1.PublicKey) -> "UserId":
        """
        Creates a user identifier from a public key.

        User identifiers are defined as the last 20 bytes of the SHA256 hash of the public key.

        Returns
        -------
            The user id.

        Example
        -------

        .. code-block:: py3

            import secp256k1

            private_key = secp256k1.PrivateKey()
            user = UserId.from_public_key(private_key.pubkey)
        """
        raw_key = public_key.serialize()

        # Sha256-hash it
        hashed_key = hashlib.sha256(raw_key).digest()

        # Keep the last 20 bytes.
        contents = hashed_key[len(hashed_key) - 20 :]
        return UserId(contents)

    def to_proto(self) -> ProtoUserId:
        """
        Convert a user identifier to its protobuf representation.
        """
        return ProtoUserId(contents=self.contents)

    @classmethod
    def from_proto(cls, proto: ProtoUserId) -> "UserId":
        """
        Create a user identifier instance from its protobuf representation.
        """
        return UserId(contents=proto.contents)

    def __str__(self) -> str:
        return self.contents.hex()


class InvalidUserId(Exception):
    """
    An exception thrown when the string representation of a user identifier is invalid
    """

    pass


@dataclass(frozen=True)
class UuidIdentifier:
    contents: bytes

    def __str__(self) -> str:
        return str(UUID(bytes=self.contents))

    def __bytes__(self) -> bytes:
        return self.contents


ProgramId = str
"""Program identifier."""

ValuesId = UUID
"""Values identifier."""

ComputeId = UUID
"""Identifier of a computation."""



import logging
from abc import ABCMeta, abstractmethod

from cosmpy.aerial.client import LedgerClient, prepare_and_broadcast_basic_transaction
from cosmpy.aerial.config import NetworkConfig
from cosmpy.aerial.tx import Transaction
from cosmpy.aerial.wallet import LocalWallet
from cosmpy.crypto.address import Address
from cosmpy.crypto.keypairs import PrivateKey
from nillion_client_proto.nillion.meta.v1.tx_pb2 import MsgPayFor, Amount

from . import Network

logger = logging.getLogger(__name__)

DEFAULT_QUERY_TIMEOUT_SECONDS = 30


class Payer(metaclass=ABCMeta):
    """
    An abstraction over the mechanism to perform payments for operations in the Nillion network.
    """

    @abstractmethod
    async def submit_payment(
        self, amount: int, resource: bytes, gas_limit: int | None = None
    ) -> str:
        """
        Submits a payment to the chain.

        This must submit a `MsgPayFor` transaction in nilchain using the given resource as a parameter.

        Arguments
        ---------
        amount
            The amount of unil that needs to be paid.

        resource
            The resource to pay for.

        gas_limit
            The gas limit to set for this operation.
        """
        pass

    @staticmethod
    def prepare_msg(resource: bytes, address: str, amount: int) -> MsgPayFor:
        """
        Create a `MsgPayFor` transaction.

        Arguments
        ---------
        resource
            The resource to pay for.

        address
            The address of the payment sender.

        amount
            The amount of unil that needs to be paid.
        """
        return MsgPayFor(
            resource=resource,
            from_address=address,
            amount=[Amount(denom="unil", amount=str(amount))],
        )


class NilChainPayer(Payer):
    """
    A payer that uses the nilchain to perform payments.
    """

    def __init__(
        self,
        network: Network,
        wallet_private_key: PrivateKey,
        gas_limit: int,
        wallet_prefix: str = "nillion",
        query_timeout_seconds: int = DEFAULT_QUERY_TIMEOUT_SECONDS,
    ):
        """
        Construct a new nilchain client.

        This allows making payments on the Nillion chain for all operations on the Nillion network.

        Arguments
        ---------
        network
            The network the payments should be made in.

        wallet_private_key
            The private key for the wallet to be used for payments.

        gas_limit
            The gas limit to apply to all payments.

        wallet_prefix
            The prefix used for addresses in the used chain.

        Example
        -------

        .. code-block:: py3

            chain_client = NilChainPayer(
                network,
                wallet_private_key=NilChainPrivateKey(bytes.fromhex(nilchain_private_key)),
                gas_limit=10000000,
            )
        """

        self.network = network
        self.payments_wallet = LocalWallet(wallet_private_key, wallet_prefix)
        self.gas_limit = gas_limit
        payments_config = NetworkConfig(
            chain_id=network.chain_id,
            url=f"grpc+{network.chain_grpc_endpoint}/",
            fee_minimum_gas_price=0,
            fee_denomination="unil",
            staking_denomination="unil",
            faucet_url=None,
        )
        self.payments_client = LedgerClient(
            payments_config, query_timeout_secs=query_timeout_seconds
        )

    @property
    def wallet_address(self) -> str:
        """
        Get the address associated with the payer's wallet.
        """
        return str(self.payments_wallet.address())

    async def submit_payment(
        self, amount: int, resource: bytes, gas_limit: int | None = None
    ) -> str:
        transaction = Transaction()
        message = NilChainPayer.prepare_msg(
            resource, str(Address(self.payments_wallet.public_key(), "nillion")), amount
        )
        transaction.add_message(message)

        gas_limit = gas_limit if gas_limit else self.gas_limit
        logger.info(
            "Submitting transaction for %s unil, resource %s", amount, resource.hex()
        )
        submitted_transaction = prepare_and_broadcast_basic_transaction(
            self.payments_client, transaction, self.payments_wallet, gas_limit=gas_limit
        )

        tx_hash = submitted_transaction.tx_hash
        logger.info("Waiting for transaction %s to be committed", tx_hash)
        submitted_transaction.wait_to_complete()

        logger.info("Transaction %s was committed", tx_hash)
        return tx_hash


class DummyPayer(Payer):
    """
    A payer that doesn't support paying.

    This can be used when a client is to be instantiated only to be used for operations that
    don't require payments.
    """

    async def submit_payment(
        self, amount: int, resource: bytes, gas_limit: int | None = None
    ) -> str:
        raise Exception("payments not supported")



from dataclasses import dataclass, field
from typing import Set, Dict, List
from nillion_client_proto.nillion.permissions.v1.permissions import (
    Permissions as ProtoPermissions,
    ComputePermissions as ProtoComputePermissions,
)
from nillion_client_proto.nillion.permissions.v1.update import (
    PermissionCommand as ProtoPermissionCommand,
    ComputePermissionCommand as ProtoComputePermissionCommand,
)
from nillion_client import UserId, ProgramId


@dataclass
class ComputePermission:
    """
    A record of compute permissions.
    """

    program_ids: Set[ProgramId] = field(default_factory=set)


@dataclass
class ComputePermissions:
    """
    A collection of compute permissions, mapping user IDs to their permissions.
    """

    permissions: Dict[UserId, ComputePermission] = field(default_factory=dict)

    def to_proto(self) -> List[ProtoComputePermissions]:
        """
        Convert this ComputePermissions instance to a list of ProtoComputePermissions messages.
        """
        return [
            ProtoComputePermissions(
                user=user_id.to_proto(), program_ids=list(permission.program_ids)
            )
            for user_id, permission in self.permissions.items()
        ]

    @classmethod
    def from_proto(
        cls, proto_list: List[ProtoComputePermissions]
    ) -> "ComputePermissions":
        """
        Create a ComputePermissions instance from a list of ProtoComputePermissions messages.
        """
        permissions = {
            UserId.from_proto(proto_perm.user): ComputePermission(
                program_ids=set(proto_perm.program_ids)
            )
            for proto_perm in proto_list
        }
        return cls(permissions)


@dataclass
class Permissions:
    """
    The permissions associated with a set of stored values.
    """

    owner: UserId
    retrieve: Set[UserId] = field(default_factory=set)
    update: Set[UserId] = field(default_factory=set)
    delete: Set[UserId] = field(default_factory=set)
    compute: ComputePermissions = field(default_factory=ComputePermissions)

    @classmethod
    def defaults_for_user(cls, user_id: UserId) -> "Permissions":
        """Returns a Permissions object with update, retrieve and delete permissions for a user."""
        retrieve = {user_id}
        update = {user_id}
        delete = {user_id}
        compute = ComputePermissions()
        return cls(
            owner=user_id,
            retrieve=retrieve,
            update=update,
            delete=delete,
            compute=compute,
        )

    def allow_retrieve(self, user_id: UserId) -> "Permissions":
        """
        Allow a user to retrieve these values.
        """
        self.retrieve.add(user_id)
        return self

    def allow_delete(self, user_id: UserId) -> "Permissions":
        """
        Allow a user to delete these values.
        """
        self.delete.add(user_id)
        return self

    def allow_update(self, user_id: UserId) -> "Permissions":
        """
        Allow a user to update these values.
        """
        self.update.add(user_id)
        return self

    def allow_compute(self, user_id: UserId, program_id: ProgramId) -> "Permissions":
        """
        Allow a user to use these values on an execution of the given program id.
        """
        if user_id not in self.compute.permissions:
            self.compute.permissions[user_id] = ComputePermission()
        self.compute.permissions[user_id].program_ids.add(program_id)
        return self

    def to_proto(self) -> ProtoPermissions:
        """
        Convert this Permissions instance to a ProtoPermissions message.
        """
        return ProtoPermissions(
            owner=self.owner.to_proto(),
            retrieve=[user_id.to_proto() for user_id in self.retrieve],
            update=[user_id.to_proto() for user_id in self.update],
            delete=[user_id.to_proto() for user_id in self.delete],
            compute=self.compute.to_proto(),
        )

    @classmethod
    def from_proto(cls, proto: ProtoPermissions) -> "Permissions":
        """
        Create a Permissions instance from a ProtoPermissions message.
        """
        return cls(
            owner=UserId.from_proto(proto.owner),
            retrieve=set([UserId.from_proto(user_id) for user_id in proto.retrieve]),
            update=set([UserId.from_proto(user_id) for user_id in proto.update]),
            delete=set([UserId.from_proto(user_id) for user_id in proto.delete]),
            compute=ComputePermissions.from_proto(proto.compute),
        )


@dataclass
class PermissionCommand:
    """A command to grant/revoke permissions"""

    grant: Set[UserId] = field(default_factory=set)
    revoke: Set[UserId] = field(default_factory=set)

    def to_proto(self) -> ProtoPermissionCommand:
        """Convert an instance to its protobuf representation"""
        return ProtoPermissionCommand(
            grant=[user.to_proto() for user in self.grant],
            revoke=[user.to_proto() for user in self.revoke],
        )

    @staticmethod
    def from_proto(proto: ProtoPermissionCommand) -> "PermissionCommand":
        """Construct an instance from its protobuf representation"""
        return PermissionCommand(
            grant=set([UserId.from_proto(user) for user in proto.grant]),
            revoke=set([UserId.from_proto(user) for user in proto.revoke]),
        )


@dataclass
class ComputePermissionCommand:
    """A command to grant/revoke compute permissions"""

    grant: ComputePermissions = field(default_factory=ComputePermissions)
    revoke: ComputePermissions = field(default_factory=ComputePermissions)

    def to_proto(self) -> ProtoComputePermissionCommand:
        """Convert an instance to its protobuf representation"""
        return ProtoComputePermissionCommand(
            grant=self.grant.to_proto(),
            revoke=self.revoke.to_proto(),
        )

    @staticmethod
    def from_proto(proto: ProtoComputePermissionCommand) -> "ComputePermissionCommand":
        """Construct an instance from its protobuf representation"""
        return ComputePermissionCommand(
            grant=ComputePermissions.from_proto(proto.grant),
            revoke=ComputePermissions.from_proto(proto.revoke),
        )


@dataclass
class PermissionsDelta:
    """A delta of permission grants/revokes to be applied"""

    retrieve: PermissionCommand = field(default_factory=PermissionCommand)
    update: PermissionCommand = field(default_factory=PermissionCommand)
    delete: PermissionCommand = field(default_factory=PermissionCommand)
    compute: ComputePermissionCommand = field(default_factory=ComputePermissionCommand)



import logging
import socket
import asyncio
from typing import Callable, Set, Type, Tuple
from grpclib import GRPCError, Status
from tenacity import (
    retry,
    wait_exponential,
    stop_any,
    stop_after_attempt,
    stop_after_delay,
    RetryCallState,
)


def _invoke_with_retry(func: Callable) -> Callable:
    max_attempts: int = 10
    max_sleep_seconds: int = 5
    timeout_seconds: int = 60

    retryable_grpc_statuses: Set[Status] = {
        Status.DEADLINE_EXCEEDED,
        Status.RESOURCE_EXHAUSTED,
        Status.UNAVAILABLE,
        Status.DATA_LOSS,
    }
    retryable_exceptions: Tuple[Type[BaseException], ...] = (
        GRPCError,
        OSError,
        ConnectionError,
        TimeoutError,
        socket.timeout,
        asyncio.TimeoutError,
    )

    def is_retryable_exception(retry_state: RetryCallState) -> bool:
        if retry_state.outcome is None or retry_state.outcome.exception() is None:
            return False

        exception = retry_state.outcome.exception()

        if not isinstance(exception, retryable_exceptions):
            return False

        if isinstance(exception, GRPCError):
            return exception.status in retryable_grpc_statuses

        return True

    def log_error(retry_state: RetryCallState) -> None:
        exception = retry_state.outcome.exception() if retry_state.outcome else None
        attempt_number: int = retry_state.attempt_number
        sleep_time: float = retry_state.upcoming_sleep or 0.0
        function_name = retry_state.fn.__name__ if retry_state.fn else "<unknown>"
        logging.warning(
            f"Invocation of {function_name} failed with error: {exception}, "
            f"attempt: {attempt_number}, "
            f"sleep_time: {sleep_time}s, "
            f"timeout: {timeout_seconds}s"
        )

    return retry(
        wait=wait_exponential(max=max_sleep_seconds),
        stop=stop_any(
            stop_after_attempt(max_attempts), stop_after_delay(timeout_seconds)
        ),
        retry=is_retryable_exception,
        reraise=True,
        after=log_error,
    )(func)
